/*
 * ProFTPD: mod_time -- allow/deny/limit operations based on times of day
 *                      and/or days of the week
 *
 * Copyright (c) 2000-2003 TJ Saunders
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * As a special exemption, TJ Saunders gives permission to link this program
 * with OpenSSL, and distribute the resulting executable, without including
 * the source code for OpenSSL in the source distribution.
 *
 * This is mod_time, contrib software for proftpd 1.2 and above.
 * For more information contact TJ Saunders <tj@castaglia.org>.
 *
 * $Id: mod_time.c,v 1.65 2004/05/12 18:52:48 tj Exp tj $
 */

#define MOD_TIME_VERSION "mod_time/2.2.1"

#include "conf.h"
#include "privs.h"

/* Make sure the version of proftpd is as necessary. */
#if PROFTPD_VERSION_NUMBER < 0x0001021001
# error "ProFTPD 1.2.10rc1 or later required"
#endif

/* just some convenient conversion factors */

#define DAYS_PER_WEEK 7
#define SECONDS_PER_HOUR 3600
#define SECONDS_PER_MINUTE 60

typedef struct {

  char *name;

  long start_time;
  long end_time;

  int days[DAYS_PER_WEEK];

  int not_hours;
  int not_days;

  char *days_arg;
  char *hours_arg;
  char *deny_mesg_file;

} pr_named_time_t;

typedef struct {

  char *targets;

  int cmdc;
  char **cmdv;

  int namec;
  char **namev;

} pr_time_map_t;

/* from src/dirtree.c */
extern xaset_t *servers;

/* for use by time_display_file(), for the %c cookie */
static char *time_curr_cmd = NULL;

/* rather than play with session.user, have my own static variable */
static char *time_user = NULL;

/* define the meta-command expansions here, rather than using the
 * G_* defines.
 */
#define DIRS_CMDC 6
static char DIRS_CMDV[DIRS_CMDC][5] =
  {"CDUP", "CWD", "LIST", "NLST", "XCWD", "XCUP"};
#define READ_CMDC 6
static char READ_CMDV[READ_CMDC][5] =
  {"MDTM", "PWD", "RETR", "SIZE", "STAT", "XPWD"};
#define WRITE_CMDC 10 
static char WRITE_CMDV[WRITE_CMDC][5] =
  {"APPE", "DELE", "MKD", "RMD", "RNFR", "RNTO", "STOR", "STOU", "XMKD", "XRMD"};

/* support function prototypes */

static void build_group_arrays(pool *, struct passwd *, char *,
  array_header **, array_header **);
int class_or_expression(char **);
static pr_time_map_t *get_class_cmd_times(cmd_rec *);
static pr_time_map_t *get_cmd_times(cmd_rec *);
static char *get_deny_mesg(cmd_rec *);
static pr_time_map_t *get_global_cmd_times(cmd_rec *);
static pr_time_map_t *get_group_cmd_times(cmd_rec *);
static pr_named_time_t *get_named_time(char *);
static pr_time_map_t *get_named_time_map(char *, char **);
static void get_system_time(long *, int *);
static pr_time_map_t *get_user_cmd_times(cmd_rec *);
static char *get_word_from_arg(char **);
int group_or_expression(char **);
static int is_allowed_time(pr_named_time_t *);
static int is_limited_cmd(cmd_rec *, pr_time_map_t *);
static char *is_valid_days_arg(char *, pr_named_time_t *);
static char *is_valid_hours_arg(char *, pr_named_time_t *);
static char *is_valid_hours_str(char *, long *);
static int is_valid_named_time(char *);
static char **parse_class_arg(char *);
static int parse_command_arg(char *, pr_time_map_t *);
static char **parse_group_arg(char *);
static char *parse_name_arg(char *, pr_time_map_t *);
static char **parse_user_arg(char *);
static config_rec *resolve_user(pool *, char **, char **, char **);
static int time_display_file(const char *, const char *);
int user_or_expression(char **);

/* Support routines
 */

static void build_group_arrays(pool *p, struct passwd *pw, char *name,
    array_header **gids, array_header **groups) {

  struct group *gr;
  array_header *xgids,*xgroups;

#ifdef HAVE_INITGROUPS
  /* Optimization for large groups, especially.
   */
  int i, total;

  i = total = 0;

  /* If we don't have a password file, we create an empty array and punt.
   */
  if (!pw) {
    if (!name || !(pw = (struct passwd *) auth_getpwnam(p, name))) {
      *gids = make_array(p, 2, sizeof(gid_t));
      *groups = make_array(p, 2, sizeof(char *));
      return;
    }
  }

  /* We need root privileges for initgroups to succeed.
   */
  PRIVS_ROOT;
  initgroups(pw->pw_name, pw->pw_gid);
  PRIVS_RELINQUISH;

  /* Allocate the appropriate space.
   */
  total = getgroups(0, 0);
  xgids = make_array(p, total, sizeof(gid_t));
  xgroups = make_array(p, 2, sizeof(char *));

  /* Populate the array of GIDs.
   */
  xgids->nelts = getgroups(total, (GETGROUPS_T *) xgids->elts);

  /* Now populate the names of the groups.
   */
  for (i = 0; i < total; i++) {
    if ((gr = (struct group *) auth_getgrgid(p, ((gid_t *) xgids->elts)[i])))
      *((char **) push_array(xgroups)) = pstrdup(p, gr->gr_name);
  }

#else /* No HAVE_INITGROUPS */

  char **gr_mem;

  /* Allocate space.
   */
  xgids = make_array(p, 2, sizeof(gid_t));
  xgroups = make_array(p, 2, sizeof(char *));

  /* Nothing there...punt.
   */
  if (!pw) {
    if (!name || !(pw = (struct passwd *) pr_auth_getpwnam(p, name))) {
      *gids = xgids;
      *groups = xgroups;
      return;
    }
  }

  /* Populate the first group name.
   */
  if ((gr = pr_auth_getgrgid(p, pw->pw_gid)) != NULL)
    *((char **) push_array(xgroups)) = pstrdup(p, gr->gr_name);

  pr_auth_setgrent(p);

  /* This is where things get slow, expensive, and ugly.
   * Loop through everything, checking to make sure we haven't already added
   * it.  This is why we have getgroups() and company.
   */
  while ((gr = pr_auth_getgrent(p)) != NULL && gr->gr_mem)
    for (gr_mem = gr->gr_mem; *gr_mem; gr_mem++) {
      if (strcmp(*gr_mem, pw->pw_name) == 0) {
        *((int *) push_array(xgids)) = (int) gr->gr_gid;
        if (pw->pw_gid != gr->gr_gid)
          *((char **) push_array(xgroups)) = pstrdup(p, gr->gr_name);
        break;
      }
    }
#endif /* HAVE_INITGROUPS */

  *gids = xgids;
  *groups = xgroups;
}

/* boolean "class-expression" matching, returns 1 if the expression
 * matches.  Code borrowed from dirtree.c's user_expression() -- unlike
 * {group,user}_expression(), though, this evaluates the list in an OR fashion,
 * not an AND fashion (a session cannot belong to more than one Class, making
 * an AND list impractical). -- TJ
 */
int class_or_expression(char **expr) {
  int found = FALSE;
  char *class = NULL;

  for (; *expr; expr++) {
    class = *expr;
    found = FALSE;

    if (class && *class && *class == '!') {
      found = !found;
      class++;
    }

    if ((class && strcmp(class, "*") == 0) ||
        (session.class && strcmp(session.class->cls_name, class) == 0))
      found = !found;

    if (!found)
      continue;

    else
      return TRUE;
  }

  return FALSE;
}

static pr_time_map_t *get_class_cmd_times(cmd_rec *cmd) {
  unsigned char *class_engine = NULL;
  pr_time_map_t *time_map = NULL;

  class_engine = get_param_ptr(main_server->conf, "Classes", FALSE);

  if (class_engine && *class_engine == TRUE) {

    /* Of necessity, the class member of the session struct _must_ be
     * initialized before evaluation of a class-expression can be done.
     * Make sure find_class() doesn't return NULL, either, else mod_time
     * will SIGSEGV.
     */
    if (session.class == NULL) {
      /* No point in continuing if no Class has been found. */
       return NULL;  
    }

  } else {
    /* No point in continuing if Classes have not been enabled. */
    return NULL;
  }

  /* Find an applicable ClassCommandTimes.  Cycle through the ClassCommandTimes,
   * first checking to see if they apply to the current class.  If so, see if
   * they apply to the current command.  If no ClassCommandTimes are found
   * apply to both the class and the command, return NULL.  Return the first
   * ClassCommandTime that applies to both the current class and command.
   */

  time_map = (pr_time_map_t *) get_param_ptr(CURRENT_CONF,
    "ClassCommandTimes", FALSE);

  while (time_map) {
    char **classes = NULL;

    /* wildcard "*" matches any class
     */
    if (*(time_map->targets) == '*' &&
        is_limited_cmd(cmd, time_map))
      return time_map;

    classes = parse_class_arg(time_map->targets);

    if (class_or_expression(classes) &&
        is_limited_cmd(cmd, time_map))
      return time_map;

    time_map = (pr_time_map_t *) get_param_ptr_next("ClassCommandTimes", TRUE);
  }

  return NULL;
}

static pr_time_map_t *get_cmd_times(cmd_rec *cmd) {
  pr_time_map_t *time_map = NULL;

  /* Check the current user for any applicable time restrictions. */
  if ((time_map = get_user_cmd_times(cmd)) != NULL)
    return time_map;

  /* If nothing found, check the current user's groups for applicable time
   * restrictions.
   */
  if ((time_map = get_group_cmd_times(cmd)) != NULL)
    return time_map;

  /* If nothing found, check the current user's class for applicable time
   * restrictions.
   */
  if ((time_map = get_class_cmd_times(cmd)) != NULL)
    return time_map;

  /* If nothing found, check for generally applicable time restrictions. */
  if ((time_map = get_global_cmd_times(cmd)) != NULL)
    return time_map;

  return NULL;
}

/* returns a string containing the filename to be sent to the client
 * containing a denial message.  A returned NULL signals that the
 * command is OK, that there are no time restrictions (thus no deny message
 * file) on the command.  An empty string will be returned in the event that
 * there is a time restriction, but no apparent denial message file
 * (ie multiple NamedTimes have been set, but no configured
 * DisplayAfterHours).
 */
static char *get_deny_mesg(cmd_rec *cmd) {
  pr_time_map_t *time_map = NULL;
  char *after_hours_mesg = NULL;
  register unsigned int i = 0;

  /* Find a time map that applies to this command. */
  if ((time_map = get_cmd_times(cmd)) != NULL) {

    /* Loop through the map's NamedTimes list, looking for the one whose time
     * matches -- first match wins.  However, does "match" mean allowed,
     * or denied?  I think it means "denied" -- search through the listed
     * NamedTimes, and deny the command if any of the hours block the command.
     */

    for (i = 0; i < time_map->namec; i++) {
      pr_named_time_t *named_time = NULL;

      /* find the matching NamedTime
       */
      if ((named_time = get_named_time((time_map->namev)[i])) == NULL) {
        log_debug(DEBUG0, MOD_TIME_VERSION
          ": error: no matching NamedTime found for '%s'",
          (time_map->namev)[i]);
        continue;
      }

      /* OK, we have a matched NamedTime.  Now check the time to see
       * if it is allowed.
       */
      if (is_allowed_time(named_time)) {
        /* Success!  Return NULL, and allow the requested action to occur
         */
        log_debug(DEBUG3, MOD_TIME_VERSION
          ": command '%s' allowed by NamedTime '%s'", cmd->argv[0],
          (time_map->namev)[i]);
        return NULL;

      } else {

        /* merely log that this NamedTime _would_ block the requested
         * command.  It doesn't -- yet -- because there may be other
         * NamedTimes in the list that do allow the command.
         */
        log_debug(DEBUG3, MOD_TIME_VERSION
          ": command '%s' denied by NamedTime '%s'", cmd->argv[0],
          (time_map->namev)[i]);
      }
    }

    /* we've reached the end of the listed NamedTimes, and none have
     * allowed the command, which means it should be denied.
     */

    /* Use the current named_time hours/days for the magic cookie
     * expansions.
     *
     * Here's how the message to displayed is found:
     *
     * 1. If there is a DisplayAfterHours, return the DisplayAfterHours
     *    filename.
     * 2. If there are no configured DisplayAfterHours, return an empty string.
     */

    if ((after_hours_mesg = (char *) get_param_ptr(session.anon_config ?
        session.anon_config->subset : main_server->conf, "DisplayAfterHours",
        FALSE)) != NULL) {

      struct stat sbuf;

      if (!pr_fsio_lstat(after_hours_mesg, &sbuf)) {
        return after_hours_mesg;

      } else {
        log_debug(DEBUG2, MOD_TIME_VERSION
          ": error finding DisplayAfterHours '%s': %s",
          after_hours_mesg, strerror(errno));
        return "";
      }

    } else {
      return "";
    }
  }

  return NULL;
}

static pr_time_map_t *get_global_cmd_times(cmd_rec *cmd) {

  pr_time_map_t *time_map = NULL;

  /* find a CommandTimes parameter; if one exists, it's applicable
   */

  /* NOTE: last arg...TRUE or FALSE (for recursion)? AND watch those
   * configuration context to search...CURRENT_CONF changes w/ commands,
   * and the CF_MERGEDOWN will mess with things, too...
   *
   * So, should I use CURRENT_CONF, and TRUE recursion, or main_server->conf,
   * and FALSE recursion?  I'm leaning towards the former, because of the
   * CF_MERGEDOWN that will probably be done.  If this is the case, though,
   * how does one reinstate a command at a lower context that has been
   * limited more strictly in a higher one?
   */

  time_map = (pr_time_map_t *) get_param_ptr(CURRENT_CONF,
    "CommandTimes", FALSE);

  while (time_map) {
    if (is_limited_cmd(cmd, time_map))
      return time_map;

    time_map = (pr_time_map_t *) get_param_ptr_next("CommandTimes",
      TRUE);
  }

  return NULL;
}

static pr_time_map_t *get_group_cmd_times(cmd_rec *cmd) {

  pr_time_map_t *time_map = NULL;

  /* retrieve the current user name
   */
  char *user = (char *) get_param_ptr(cmd->server->conf,
       C_USER, FALSE);

  /* of necessity, the group members of the session struct _must_ be
   * initialized before evaluation of a group-expression can be done.
   */
  if (session.gids == NULL || session.groups == NULL)
    build_group_arrays(cmd->pool, NULL, user, &session.gids,
      &session.groups);

  /* Find an applicable GroupCommandTime.  Cycle through the GroupCommandTimes,
   * first checking to see if they apply to the current group.  If so, see if
   * they apply to the current command.  If no GroupCommandTimes are found
   * apply to both the group and the command, return NULL.  Return the first
   * GroupCommandTime that applies to both the current user and command.
   */

  time_map = (pr_time_map_t *) get_param_ptr(CURRENT_CONF,
    "GroupCommandTimes", FALSE);

  while (time_map) {
    char **groups = NULL;

    /* wildcard "*" matches any group
     */
    if (*(time_map->targets) == '*' &&
        is_limited_cmd(cmd, time_map))
      return time_map;

    groups = parse_group_arg(time_map->targets);

    if (pr_group_and_expression(groups) &&
        is_limited_cmd(cmd, time_map))
      return time_map;

    time_map = (pr_time_map_t *) get_param_ptr_next("GroupCommandTimes", TRUE);
  }

  return NULL;
}

static pr_named_time_t *get_named_time(char *requested_name) {
  config_rec *c = NULL;
  pr_named_time_t *named_time = NULL;

  /* check through all the NamedTimes, looking for the given name
   */
  c = find_config(main_server->conf, CONF_PARAM, "NamedTime", FALSE);

  while (c) {
    named_time = (pr_named_time_t *) c->argv[0];

    if (strcmp(named_time->name, requested_name) == 0)
      return named_time;

    c = find_config_next(c, c->next, CONF_PARAM, "NamedTime", FALSE);
  }

  /* default return value
   */
  return NULL;
}

static pr_time_map_t *get_named_time_map(char *name_arg, char **unknown_name) {
  pr_time_map_t *time_map = NULL;

  time_map = (pr_time_map_t *) pcalloc(permanent_pool,
    sizeof(pr_time_map_t));

  if ((*unknown_name = parse_name_arg(name_arg, time_map)) != NULL)
    return NULL;

  return time_map;
}

static void get_system_time(long *current_time, int *current_day_of_week) {
  unsigned char times_gmt = TRUE, *tmp = NULL;

  /* Retrieve the current system time of day. */
  time_t system_time;
  struct tm *timep = NULL;

  /* Get the system's time.  */
  system_time = time(NULL);

  /* If the "TimesGMT" directive is in effect... */
  if ((tmp = get_param_ptr(TOPLEVEL_CONF, "TimesGMT", FALSE)))
    times_gmt = *tmp;
  
  if (times_gmt) {

    /* ...convert the time using gmtime()... */

    if ((timep = gmtime(&system_time)) == NULL) {

      /* If GMT is not available, gmtime() will return NULL.  If this
       * is the case, default to using localtime(), and report the
       * problem.
       */
      log_pri(LOG_INFO, "gmtime() returned NULL -- using localtime()");
      timep = localtime(&system_time);
    }

  } else {

    /* ...convert it using localtime() */
    timep = localtime(&system_time);
  }

  /* Calculate the current time, in seconds. */

  *current_time = (
    ((long) timep->tm_hour * SECONDS_PER_HOUR) +
    ((long) timep->tm_min * SECONDS_PER_MINUTE) +
    (long) timep->tm_sec);

  /* don't forget to return the day of the week, if requested
   */

  if (current_day_of_week != NULL)
    *current_day_of_week = timep->tm_wday;

  return;
}

static pr_time_map_t *get_user_cmd_times(cmd_rec *cmd) {
  pr_time_map_t *time_map = NULL;
  config_rec *c = NULL;
  struct passwd *pw;

  /* Retrieve the current user name. */
  char *user = (char *) get_param_ptr(cmd->server->conf,
    C_USER, FALSE);

  /* Use this knowledge to populate the necessary members of the
   * session struct.
   *
   * Need to check if the authenticating user is an UserAlias.  If it is an
   * UserAlias, lookup the aliased user, and use that name. 
   */

  c = find_config(main_server->conf, CONF_PARAM, "UserAlias", TRUE);

  if (c) do {

    if (strcmp((c->argv)[0], "*") == 0 ||
        strcmp((c->argv)[0], user) == 0) {

      /* Replace the "logging in as" user with the real system userid
       * given by the UserAlias directive.
       *
       * UserAlias config_rec->argv[0] = alias
       *           config_rec->argv[1] = real user
       */
      user = c->argv[1];
      break;
    }

  } while ((c = find_config_next(c, c->next, CONF_PARAM, "UserAlias",
    TRUE)) != NULL);

  /* If this field hasn't been set, do a lookup and store the resultu
   * locally.
   */
  if (!session.user) {
    /* Populate the pw struct, check for a NULL return value. */

    if ((pw = pr_auth_getpwnam(cmd->tmp_pool, user)) == NULL)
      return NULL;

    time_user = pstrdup(session.pool, pw->pw_name);
  }

  /* Find an applicable UserCommandTime.  Cycle through the UserCommandTimes,
   * first checking to see if they apply to the current user.  If so, see if
   * they apply to the current command.  If no UserCommandTimes are found
   * apply to both the user and the command, return NULL.  Return the first
   * UserCommandTime that applies to both the current user and command.
   */

  time_map = (pr_time_map_t *) get_param_ptr(CURRENT_CONF,
    "UserCommandTimes", FALSE);

  while (time_map) {
    char **users = NULL;

    /* Check the "targets" field, a user OR expression in this case.
     * Wildcard "*" matches any user, always
     */

    if (*(time_map->targets) == '*' &&
        is_limited_cmd(cmd, time_map))
      return time_map;

    users = parse_user_arg(time_map->targets);

    if (user_or_expression(users) &&
        is_limited_cmd(cmd, time_map))
      return time_map;

    time_map = (pr_time_map_t *) get_param_ptr_next("UserCommandTimes", TRUE);
  }

  return NULL;
}

/* code borrowed from src/dirtree.c's get_word() -- modified to separate
 * words on commas as well as spaces. -- TJ
 */
static char *get_word_from_arg(char **arg) {

  char *ret, *dst;
  char quote_mode = 0;

  if (!arg || !*arg || !**arg)
    return NULL;

  while (**arg && isspace((unsigned char) **arg))
    (*arg)++;

  if (!**arg)
    return NULL;

  ret = dst = *arg;

  if (**arg == '\"') {
    quote_mode++;
    (*arg)++;
  }

  while (**arg && **arg != ',' &&
      (quote_mode ? (**arg != '\"') : (!isspace((unsigned char) **arg)))) {

    if (**arg == '\\' && quote_mode) {

      /* escaped char */
      if (*((*arg) + 1))
        *dst = *(++(*arg));
    }

    *dst++ = **arg;
    ++(*arg);
  }

  if (**arg)
    (*arg)++;

  *dst = '\0';

  return ret;
}

/* boolean "group-expression" matching, returns 1 if the expression
 * matches.  Code borrowed from dirtree.c's user_expression() -- unlike
 * {group,user}_expression(), though, this evaluates the list in an OR fashion,
 * not an AND fashion.
 */
int group_or_expression(char **expr) {
  register unsigned int i, found;
  char *group = NULL;

  for (; *expr; expr++) {
    group = *expr;
    found = FALSE;

    if (*group == '!') {
      found = !found;
      group++;
    }

    /* NOTE: _this_ comparison is little trickier, since the session user
     *  can belong to multiple groups, each of which must be checked.
     */
    for (i = session.groups->nelts - 1; i >= 0; i--) {
      char *current_group = *(((char **) session.groups->elts) + i);

      if (strcmp(current_group, "*") == 0 ||
          strcmp(current_group, group) == 0) {
        found = !found;
        break;
      }
    }

    if (strcmp(group, "*") == 0 ||
        strcmp(session.user ? session.user : time_user, group) == 0)
      found = !found;

    if (!found)
      continue;

    else
      return TRUE;
  }

  return FALSE;
}

static int is_allowed_time(pr_named_time_t *named_time) {
  int allowed_day = FALSE, allowed_hour = FALSE;
  long current_time;
  int current_day;

  /* Retrieve the current time and day. */
  get_system_time(&current_time, &current_day);

  /* Check the allowed days, if set, first. */ 
  if ((named_time->days)[0] != -1) {
    register unsigned int i;

    for (i = 0; i < DAYS_PER_WEEK; i++)
      if (current_day == (named_time->days)[i]) {
        allowed_day = TRUE;
        break;
      }

    if (!allowed_day && !named_time->not_days)
      return FALSE;

    if (allowed_day && named_time->not_days)
      return FALSE;
  }

  /* Check the allowed times now. */
  if (current_time >= named_time->start_time &&
      current_time <= named_time->end_time)
    allowed_hour = TRUE;

  if (!allowed_hour && !named_time->not_hours)
    return FALSE;

  if (allowed_hour && named_time->not_hours)
    return FALSE;

  /* If it's reached this point, it's an allowable time. */
  return TRUE;  
}

static int is_limited_cmd(cmd_rec *cmd, pr_time_map_t *time_map) {
  register unsigned int i = 0;

  /* Loop through the limited FTP commands listed in the time map, looking
   * for a match with the current command.
   */
  for (i = 0; i < time_map->cmdc; i++) {

    /* Check for the wildcard in the cmdv first. */
    if (strcmp((time_map->cmdv)[i], "*") == 0)
      return TRUE;

    /* Next, check for NONE.  If present, it matches any command as FALSE. */
    if (strcmp((time_map->cmdv)[i], "NONE") == 0)
      return FALSE;

    /* Treat PASS command as a special case, as it matches the "LOGIN"
     * meta-command.
     */
    if (strcmp((time_map->cmdv)[i], "LOGIN") == 0 &&
        strcmp((cmd->argv)[0], "PASS") == 0)
      return TRUE;

    /* All other meta-commands should be expanded properly, so do a
     * normal comparison.
     */
    if (strcmp((time_map->cmdv)[i], (cmd->argv)[0]) == 0)
      return TRUE;
  }

  /* Default return value -- command not limited by given time map. */
  return FALSE;
}

static char *is_valid_days_arg(char *days_arg, pr_named_time_t *named_time) {
  register int i = 0, ibuf = 0, idays = 0;
  char day_buffer[1024] = {'\0'};
  int prev_digit;

  /* As per the man page on time functions, the days of the week are given
   * integers in the following manner:  Sunday is the first day of the week
   * (0, and 7) through Saturday (6), as per crontab(5) entries.
   */

  char *extra_char_error = "extra \",\" or \"-\" detected";
  char *format_error = "day numbers must be 0-7";
  char *illegal_char_error = "unsupported char detected"; 
  char *missing_days_error = "no days specified";
  char *order_error = "day numbers must be in ascending order";
  char *wildcard_error = "bad wildcard argument";

  /* Initialize the "not days (check)" member of named_time to FALSE
   * by default.
   */
  named_time->not_days = FALSE;

  /* Initialize the prev_digit variable, used to ensure that the given
   * numbers are in ascending order.  Checks prior to this function call
   * (at least within this module) ensure that the first character of
   * days_arg will be a digit, "!", or "*".  If it's the wildcard, and _only_
   * the wildcard, then this function call is not necessary.
   */

  if (days_arg[i] == '*') {
    if ((int) strlen(days_arg) == 1) {

      /* Store an appropriate days_arg string. */
      named_time->days_arg = pstrdup(permanent_pool, "every day");
      return NULL;

    } else
      return wildcard_error;

  } else if (days_arg[i] == '!') {

    if (strlen(days_arg) < 2)
      return missing_days_error;

    named_time->not_days = TRUE;
    i++;
  }

  /* Copy the days arg for later reference. */
  named_time->days_arg = pstrdup(permanent_pool, &(days_arg[i]));

  prev_digit = atoi((char *) &(days_arg[i])); 

  if (prev_digit < 0 || prev_digit > 7)
    return format_error;

  day_buffer[ibuf++] = days_arg[i++];

  /* First pass -- check for illegal characters, and strip out whitespace. */
  for (; i < (int) strlen(days_arg); i++) {

    if (isspace((int) days_arg[i]))
      continue;

    if (!isdigit((int) days_arg[i]) &&
        days_arg[i] != ',' &&
        days_arg[i] != '-')
      return illegal_char_error;

    if (isdigit((int) days_arg[i])) {
      int digit = atoi((char *) &(days_arg[i]));

      if (digit <= prev_digit)
        return order_error;

      else
        prev_digit = digit;
    }

    if (prev_digit < 0 || prev_digit > 7)
      return format_error;

    day_buffer[ibuf++] = days_arg[i];
  }

  /* Second pass -- populate the named_time's days array member. */
  for (i = 0; i < (int) strlen(day_buffer); i++) {
    int digit;

    if (!isdigit((int) day_buffer[i]))
      return extra_char_error;

    digit = atoi((char *) &(day_buffer[i]));

    /* As per the crontab(5) man page, 0 and 7 (Sunday) are treated
     * the same; hence if the current digit (day-of-week) is 7, automatically
     * set it to 0.
     */
    if (digit == 7)
      digit = 0;

    /* If the next character is a comma, put the digit into the
     * days array, and increment the index to the next digit
     */
    if (day_buffer[i + 1] == ',') {

      /* Ensure that the next char after the comma is indeed a digit.
       * If not, it's an error.
       */
      if (!isdigit((int) day_buffer[i + 2]))
        return extra_char_error;

      (named_time->days)[idays++] = digit;
      i += 1;
      continue;

    /* If the next character is a hyphen, it signals a day range,
     * so populate the days array with the in-between numbers, up
     * to and including the end day digit.  Don't forget to increment
     * the index to the next digit for processing.
     */
    } else if (day_buffer[i + 1] == '-') {
      int next_digit = 0;
      int range_index;

      /* Ensure that the next char after the hyphen is indeed a digit.
       * If not, it's an error.
       */
      if (!isdigit((int) day_buffer[i + 2]))
        return extra_char_error;

      else
        next_digit = atoi((char *) &(day_buffer[i + 2]));

      for (range_index = digit; range_index <= next_digit; range_index++)
        (named_time->days)[idays++] = range_index;

      i += 3;
      continue;
  
    /* If the next character is anything else (except a NUL), assume it's
     * an illegal character.  Ideally, this case will never be reached,
     * but one can never tell, and one shouldn't leave a dangling
     * if statement.
     */
    } else if (day_buffer[i + 1] != '\0')
      return illegal_char_error;

    else
      (named_time->days)[idays++] = digit;
  }

  return NULL;
}

static char *is_valid_hours_arg(char *hours_arg, pr_named_time_t *named_time) {

  /* Allocate a char buffer for the time arg.  The time arg may contain
   * spaces (use of quotes in a config file argument preserves whitespace)
   * which will need to be ignored.  Also, the hours, minutes, and seconds
   * can (and will) be determined separately for the start time and end
   * time.
   */

  char start_time[1024] = {'\0'};
  char end_time[1024] = {'\0'};
  char *error = NULL;
  int flag = FALSE;
  register unsigned int i = 0, ibuf = 0;

  char *extra_char_error = "extra \":\" or \"-\" detected";
  char *illegal_char_error = "unsupported char detected";
  char *missing_hours_error = "no hours specified";
  char *order_error = "end time must be AFTER start time";

  /* Initialize the "not hours (check)" member of named_time to FALSE
   * by default.
   */
  named_time->not_hours = FALSE;

  if (*hours_arg == '!') {

    if (strlen(hours_arg) < 2)
      return missing_hours_error;

    named_time->not_hours = TRUE;
    i++;
  }

  /* Copy the hours arg for later reference. Note that session.pool has
   * not yet been assigned, so we use permanent_pool.
   */
  named_time->hours_arg = pstrdup(permanent_pool, &(hours_arg[i]));
  
  /* First pass -- check for illegal characters, and strip out whitespace. */
  for (; i < strlen(hours_arg); i++) {

    if (isspace((int) hours_arg[i]))
      continue;

    if (!isdigit((int) hours_arg[i]) &&
        hours_arg[i] != ':' &&
        hours_arg[i] != '-')
      return illegal_char_error;

    /* Toggle the flag indicating which buffer to populate, start_time
     * or end_time.  Don't forget to reset the ibuf variable, too!
     */

    else if (hours_arg[i] == '-') {
      if (flag)
        return extra_char_error;

      else {
        flag = TRUE;
        ibuf = 0;
        continue;
      }
    }

    if (flag)
      end_time[ibuf++] = hours_arg[i];

    else
      start_time[ibuf++] = hours_arg[i];
  }

  if ((error = is_valid_hours_str(start_time,
      &(named_time->start_time))) != NULL)
    return error;

  if ((error = is_valid_hours_str(end_time,
      &(named_time->end_time))) != NULL)
    return error;

  /* Make sure the end_time is _after_ the start_time. */
  if (named_time->end_time <= named_time->start_time)
    return order_error;
  
  return NULL;
}

static char *is_valid_hours_str(char *time_str, long *timeno) {

  char hours_str[3] = { "\0\0\0" };
  char minutes_str[3] = { "\0\0\0" };
  char seconds_str[3] = { "\0\0\0" };

  char *format_error = "improperly formatted time argument";
  char *hours_error = "hours cannot be greater than 23";
  char *minutes_error = "minutes cannot be greater than 59";
  char *seconds_error = "seconds cannot be greater than 59";  

  long hours, minutes, seconds;

  /* Check the given string for the proper formatting I need to determine
   * allowable start and end times against the system clock.
   */
  if (sscanf(time_str, "%2[0-9]:%2[0-9]:%2[0-9]",
      hours_str, minutes_str, seconds_str) < 3)
    return format_error;

  /* Convert the scanned in number strings to _real_ numbers. */

  hours = strtol(hours_str, NULL, 10);
  minutes = strtol(minutes_str, NULL, 10);
  seconds = strtol(seconds_str, NULL, 10);

  /* Make sure that all numbers are in the proper ranges. */

  if (hours < 0 ||  hours > 23)
    return hours_error;

  if (minutes < 0 || minutes > 59)
    return minutes_error;

  /* NOTE: this seconds is, by definition, allowed to be 61 at most,
   * for leap seconds (cf ctime(3)).  For the sake of common sense, I'll
   * limit it to 59 or less.
   */

  if (seconds < 0 || seconds > 59)
    return seconds_error;

  /* Calculate the time in seconds. */
  *timeno = ((hours * SECONDS_PER_HOUR) + (minutes * SECONDS_PER_MINUTE) +
    seconds);

  return NULL;
}

static int is_valid_named_time(char *name) {

  pr_named_time_t *named_time = NULL;

  /* check through all the NamedTimes, making sure the given named
   * set is defined.
   */
  named_time = (pr_named_time_t *) get_param_ptr(main_server->conf,
    "NamedTime", FALSE);

  while (named_time) {
    if (strcmp(named_time->name, name) == 0)
      return TRUE;

    named_time = (pr_named_time_t *) get_param_ptr_next("NamedTime", TRUE);
  }

  return FALSE;
}

/* Code borrowed from src/dirtree.c's get_config_cmd() -- modified to
 * operate only on a single string of an input mod_time configuration
 * directive
 */
static char **parse_class_arg(char *class_arg) {
  char *class = NULL, **class_arr = NULL;
  array_header *tmp_arr;

  /* Allocate an array. */
  tmp_arr = make_array(permanent_pool, 4, sizeof(char **));

  /* Add each class to the array. */
  while((class = get_word_from_arg(&class_arg)) != NULL) {
    char *tmp = pstrdup(permanent_pool, class);

    /* Push the class into the tmp array. */
    *((char **) push_array(tmp_arr)) = tmp;
  }

  /* Terminate the temp array with a NULL, as is proper. */
  *((char **) push_array(tmp_arr)) = NULL;

  class_arr = (char **) tmp_arr->elts;

  /* Return the array of classes. */
  return class_arr;
}

/* Code borrowed from src/dirtree.c's get_config_cmd() -- modified to
 * operate only on a single string of an input mod_time configuration
 * directive
 */
static int parse_command_arg(char *cmd_arg, pr_time_map_t *time_map) {
  char *cmd = NULL, *cmd_arg_dup = NULL;
  array_header *tmp_arr;
  int cmdc = 0;

  /* make a copy of the cmd_arg, and operate on that
   */
  cmd_arg_dup = pstrdup(main_server->pool, cmd_arg);

  /* Allocate an array
   * BUG: *never* allocate directly from permanent_pool!
   */
  tmp_arr = make_array(permanent_pool, 4, sizeof(char **));

  /* Add each command to the array, silently skipping USER, PASS, and QUIT
   * if they are present.  To be pedantic, I'll also check that if the
   * ALL or NONE meta-commands are given, they are the _only_ commands used,
   * so as to prevent misconfiguration problems.
   *
   * NOTE: Also check for duplicate entries, ie a LIST and a DIRS both given
   * as configuration parameters; not this is a real problem, as the first
   * duplicate in the list would match the command, but, for efficiency,
   * it would be best to catch those duplicates here.
   */
  while((cmd = get_word_from_arg(&cmd_arg_dup)) != NULL) {
    char *tmp = NULL;
    cmdc++;

    /* skip USER, PASS, and QUIT
     */
    if (strcmp(cmd, C_USER) == 0 ||
        strcmp(cmd, C_PASS) == 0 ||
        strcmp(cmd, C_QUIT) == 0)
      continue;

    if (strcmp(cmd, "ALL") == 0) {

      /* If this isn't the only command given (cmdc > 1), error out. */
      if (cmdc > 1)
        return FALSE;

      /* Expand the ALL meta-command. */
      tmp = pstrdup(permanent_pool, "*");
      *((char **) push_array(tmp_arr)) = tmp;

      /* Don't forget to increment the cmdc counter of the given time map. */
      if (time_map) {
        time_map->cmdc = tmp_arr->nelts;;
        time_map->cmdv = (char **) tmp_arr->elts;
      }

      return TRUE;
    }

    if (strcmp(cmd, "DIRS") == 0) {
      register unsigned int i;

      /* Expand the DIRS meta-command to all the FTP commands that deal
       * directories.
       */
      for (i = 0; i < DIRS_CMDC; i++) {
        tmp = pstrdup(permanent_pool, DIRS_CMDV[i]);
        *((char **) push_array(tmp_arr)) = tmp;
      }

      continue;
    }

    if (strcmp(cmd, "LOGIN") == 0) {

      /* Expand the LOGIN meta-command. */
      tmp = pstrdup(permanent_pool, "LOGIN");
      *((char **) push_array(tmp_arr)) = tmp;

      continue;
    }

    if (strcmp(cmd, "NONE") == 0) {

      /* If this isn't the only command given (cmdc > 1), error out. */
      if (cmdc > 1) 
        return FALSE;

      /* Insert the NONE "meta-command" into the list. */
      tmp = pstrdup(permanent_pool, "NONE");
      *((char **) push_array(tmp_arr)) = tmp;

      if (time_map) {
        time_map->cmdc = tmp_arr->nelts;;
        time_map->cmdv = (char **) tmp_arr->elts;
      }

      return TRUE;
    }

    if (strcmp(cmd, "READ") == 0) {
      register unsigned int i;

      /* Expand the READ meta-command to all the FTP commands that deal
       * with file reading (directory listing not included).
       */

      for (i = 0; i < READ_CMDC; i++) {
        tmp = pstrdup(permanent_pool, READ_CMDV[i]);
        *((char **) push_array(tmp_arr)) = tmp;
      }
  
      continue;
    }

    if (strcmp(cmd, "WRITE") == 0) {
      register unsigned int i;

      /* Expand the WRITE meta-command to all the FTP commands that deal
       * with file or directory write/creation/deletion (MKD and RMD included)
       */

      for (i = 0; i < WRITE_CMDC; i++) {
        tmp = pstrdup(permanent_pool, WRITE_CMDV[i]);
        *((char **) push_array(tmp_arr)) = tmp;
      }

      continue;
    }

    /* Default: just copy the command as is. */
    tmp = pstrdup(permanent_pool, cmd);
    *((char **) push_array(tmp_arr)) = tmp;
  }

  /* Populate the given time map's cmdv array w/ the parsed commands.
   * Increment the cmdc counter accordingly.
   */  
  if (time_map) {
    time_map->cmdc = tmp_arr->nelts;
    time_map->cmdv = (char **) tmp_arr->elts;
  }

  return TRUE;
}

static char **parse_group_arg(char *group_arg) {
  return parse_class_arg(group_arg);
}

static char *parse_name_arg(char *name_arg, pr_time_map_t *time_map) {
  char *name = NULL;
  array_header *tmp_arr;

  /* Allocate an array. */
  tmp_arr = make_array(permanent_pool, 4, sizeof(char **));

  /* Add each name to the array, after checking to see that it is a
   * valid (ie already configured) NamedTime name
   */
  while((name = get_word_from_arg(&name_arg)) != NULL) {
    char *tmp = NULL;
 
    /* Make sure the given name matches a configured NamedTime. */
    if (!is_valid_named_time(name))
      return name;

    tmp = pstrdup(permanent_pool, name);
 
    /* Push the NamedTime name into the tmp array. */
    *((char **) push_array(tmp_arr)) = tmp;

    /* Don't forget to increment the namec counter of the given time map. */
    if (time_map)
      time_map->namec++;
  }

  /* Terminate the temp array with a NULL, as is proper. */
  *((char **) push_array(tmp_arr)) = NULL;
 
  /* Populate the given time map's namev array w/ the parsed names. */   
  if (time_map)
    time_map->namev = (char **) tmp_arr->elts;

  return NULL;
}

/* Modified from parse_group_arg(), in order to replace any UserAliases
 */
static char **parse_user_arg(char *user_arg) {
  char *user = NULL, **user_arr = NULL;
  array_header *tmp_arr;

  /* Allocate an array. */
  tmp_arr = make_array(permanent_pool, 4, sizeof(char **));

  /* Add each user to the array. */
  while((user = get_word_from_arg(&user_arg)) != NULL) {
    config_rec *c = NULL; 
    char *tmp = NULL;

    c = find_config(main_server->conf, CONF_PARAM, "UserAlias", TRUE);

    if (c) do {
      if (strcmp(c->argv[0], "*") == 0 ||
          strcmp(c->argv[0], user) == 0) {

        /* Replace the "logging in as" user with the real system userid
         * given by the UserAlias directive.
         */
        user = c->argv[1];
        break;
      }

    } while ((c = find_config_next(c, c->next, CONF_PARAM, "UserAlias",
      TRUE)) != NULL);

    tmp = pstrdup(permanent_pool, user);

    /* Push the _real_ user into the tmp array. */
    *((char **) push_array(tmp_arr)) = tmp;
  }

  /* Terminate the temp array with a NULL, as is proper. */
  *((char **) push_array(tmp_arr)) = NULL;

  user_arr = (char **) tmp_arr->elts;

  /* Return the array of users. */
  return user_arr;
}

static config_rec *resolve_user(pool *p, char **user, char **ournamep,
    char **anonnamep) {
  config_rec *conf, *top_conf;
  char *ourname,*anonname = NULL;
  int is_alias = 0, force_anon = 0;

  /* Precendence rules:
   *   1. Search for UserAlias directive.
   *   2. Search for Anonymous directive.
   *   3. Normal user login
   */

  ourname = get_param_ptr(main_server->conf, "UserName", FALSE);

  if (ournamep && ourname)
    *ournamep = ourname;

  conf = find_config(main_server->conf, CONF_PARAM, "UserAlias", TRUE);

  if (conf) do {
    if (strcmp(conf->argv[0], "*") == 0 ||
        strcmp(conf->argv[0], *user) == 0) {
      is_alias = 1;
      break;
    } 

  } while ((conf = find_config_next(conf, conf->next, CONF_PARAM,
    "UserAlias", TRUE)) != NULL);

  /* if AuthAliasOnly is set, ignore this one and continue */
  top_conf = conf;

  while (conf && conf->parent &&
        find_config(conf->parent->set, CONF_PARAM, "AuthAliasOnly", FALSE)) {

    is_alias = 0;
    find_config_set_top(top_conf);
    conf = find_config_next(conf, conf->next, CONF_PARAM, "UserAlias", TRUE);

    if (conf && (!strcmp(conf->argv[0], "*") || !strcmp(conf->argv[0], *user)))
      is_alias = 1;
  }

  if (conf) {
    *user = conf->argv[1];

    /* If the alias is applied inside an <Anonymous> context, we have found
     * our anon block.
     */
    if (conf->parent && conf->parent->config_type == CONF_ANON)
      conf = conf->parent;

    else
      conf = NULL;
  }

  /* Next, search for an anonymous entry */

  if (!conf)
    conf = find_config(main_server->conf, CONF_ANON, NULL, FALSE);

  else
    find_config_set_top(conf);

  if (conf) do {
    anonname = get_param_ptr(conf->subset, "UserName", FALSE);

    if (!anonname)
      anonname = ourname;

    if (anonname && !strcmp(anonname, *user)) {

      if (anonnamep)
        *anonnamep = anonname;
      break;
    }

  } while ((conf = find_config_next(conf, conf->next, CONF_ANON, NULL,
    FALSE)) != NULL);

  if (!is_alias && !force_anon) {

    if (find_config((conf ? conf->subset :
        main_server->conf), CONF_PARAM, "AuthAliasOnly", FALSE)) {

      if (conf && conf->config_type == CONF_ANON)
        conf = NULL;

      else
        *user = NULL;

      if (*user && find_config(main_server->conf, CONF_PARAM, "AuthAliasOnly",
          FALSE))
        *user = NULL;

      if ((!user || !conf) && anonnamep)
        *anonnamep = NULL;
    }
  }

  return conf;
}

/* copied liberally from mod_core's core_display_file(), except that this
 * uses add_response_err() rather than send_response_raw().
 */
static int time_display_file(const char *numeric, const char *filename) {
   int is_first_line = TRUE, *current_clients = NULL, classes_enabled = 0;
   pr_fh_t *mesg_fh = NULL;
   char buffer[1024] = {'\0'};
   unsigned int *max_clients = NULL;
   unsigned char *class_engine = NULL;
   pool *p = make_sub_pool(session.pool);
   xaset_t *current_set = NULL;
 
   /* Display strings. */
   const char *mesg_time = NULL;
   char *mesg_line = NULL;
   char mesg_max[12] = "unlimited";
   char mesg_cur[12] = {'\0'}, mesg_cur_class[12] = {'\0'};
   char config_class_users[128] = {'\0'}, *user = NULL;
 
   /* Open the deny message file. */
   mesg_fh = pr_fsio_open_canon(filename, O_RDONLY);
   if (mesg_fh == NULL)
    return -1;
 
   current_set = (session.anon_config ? session.anon_config->subset :
     main_server->conf);
 
   /* Get the current time. */
   mesg_time = pr_strtime(time(NULL));
 
   /* Get the MaxClients limit. */
   max_clients = get_param_ptr(current_set, "MaxClients", FALSE);
 
   /* Get the number of current connections. */
   current_clients = get_param_ptr(main_server->conf, "CURRENT-CLIENTS", FALSE);

   snprintf(mesg_cur, sizeof(mesg_cur), "%d",
     current_clients ? *current_clients : 1);
 
   /* Check to see if Classes have been enabled. */
   class_engine = get_param_ptr(TOPLEVEL_CONF, "Classes", FALSE);

   if (class_engine && *class_engine == TRUE)
     classes_enabled = 1;
   
   if (classes_enabled && session.class && session.class->cls_name) {
     int *class_users = NULL;

     snprintf(config_class_users, sizeof(config_class_users),
       "CURRENT-CLIENTS-CLASS-%s", session.class->cls_name);

     class_users = get_param_ptr(main_server->conf, config_class_users, FALSE);

     snprintf(mesg_cur_class, sizeof(mesg_cur_class), "%d",
       class_users ? *class_users : 0);

   } else {
      mesg_cur_class[0] = 0;
   }
 
   /* Resolve the user. */
   if ((user = (char *) get_param_ptr(main_server->conf, C_USER,
       FALSE)) == NULL)
     user = "";
 
   /* Read out and send the deny message, line by line, from the file. */
   while (pr_fsio_gets(buffer, sizeof(buffer), mesg_fh) != NULL) {
 
     int buffer_len;
 
     /* Make sure the buffer is NUL-terminated. */
     buffer[sizeof(buffer)-1] = '\0';
 
     /* Get the buffer length. */
     buffer_len = strlen(buffer);
 
     /* *Sigh* Deal w/ newline/carriage returns. */
     while (buffer_len &&
           (buffer[buffer_len - 1] == '\r' ||
            buffer[buffer_len - 1] == '\n')) {
       buffer[buffer_len - 1] = '\0';
       buffer_len--;
     }
 
     /* Perform the custom variable substitution. */
     mesg_line = sreplace(p, buffer,
       "%T", mesg_time,
       "%C", (session.cwd[0] ? session.cwd : "(none)"),
       "%c", time_curr_cmd,
       "%R", (session.c && session.c->remote_name ?  session.c->remote_name : "(unknown)"),
       "%L", main_server->ServerFQDN,
       "%U", user,
       "%u", session.user,
       "%M", mesg_max,
       "%N", mesg_cur,
       "%E", main_server->ServerAdmin,
       "%V", main_server->ServerName,
       "%x", (classes_enabled && session.class) ? session.class->cls_name : "",
       "%y", mesg_cur_class,
       "%z", "-", NULL);
 
     if (is_first_line) {
       pr_response_add_err(numeric, "%s", mesg_line);
       is_first_line = FALSE;
 
     } else
       pr_response_add_err(R_DUP, "%s", mesg_line);
   }
 
   pr_fsio_close(mesg_fh);
   return 0;
}

/* Boolean "user-expression" matching, returns 1 if the expression
 * matches.  Code borrowed from dirtree.c's user_expression() -- unlike
 * {group,user}_expression(), though, this evaluates the list in an OR fashion,
 * not an AND fashion (a session cannot "be" to more than one user, making
 * an AND list impractical). -- TJ
 */
int user_or_expression(char **expr) {
  int found;
  char *user;

  for (; *expr; expr++) {
    user = *expr;
    found = FALSE;

    if (*user == '!') {
      found = !found;
      user++;
    }

    if (strcmp(user, "*") == 0 ||
        strcmp(session.user ? session.user : time_user, user) == 0)
      found = !found;

    if (!found)
      continue;

    else
      return TRUE;
  }

  return FALSE;
}

/* Configuration directive handlers
 */

MODRET add_namedtime(cmd_rec *cmd) {
  pr_named_time_t *named_time = NULL;
  register unsigned int i;
  char *error = NULL;
  config_rec *c = NULL;

  CHECK_VARARGS(cmd, 3, 4);
  CHECK_CONF(cmd, CONF_ROOT);

  /* Check for NamedTimes of the requested name. */
  named_time = (pr_named_time_t *) get_param_ptr(main_server->conf,
    "NamedTime", FALSE);

  while (named_time) {
    if (strcmp(named_time->name, (cmd->argv)[1]) == 0)
      return ERROR_MSG(cmd, NULL, pstrcat(cmd->tmp_pool,
        (cmd->argv)[0], ": requested NamedTime '",
        (cmd->argv)[1], "' already defined", NULL));

    named_time = (pr_named_time_t *) get_param_ptr_next("NamedTime", TRUE);
  }

  named_time = (pr_named_time_t *) pcalloc(permanent_pool,
    sizeof(pr_named_time_t));

  /* Initialize some of the fields of the newly-allocated NamedTime struct. */
  named_time->days_arg = NULL;
  named_time->hours_arg = NULL;
  named_time->deny_mesg_file = NULL;
  
  /* Initialize the days member of the named_time struct. */
  for (i = 0; i < DAYS_PER_WEEK; i++)
    (named_time->days)[i] = -1;

  /* Check the time argument. */
  if ((cmd->argv)[2][0] == '!') {

    if (!isdigit((int) (cmd->argv)[2][1]) ||
        !isdigit((int) (cmd->argv)[2][strlen((cmd->argv)[2]) - 1]))
      CONF_ERROR(cmd,
        "please begin and end hours argument with a number");

  } else {
  
    if (!isdigit((int) (cmd->argv)[2][0]) ||
        !isdigit((int) (cmd->argv)[2][strlen((cmd->argv)[2]) - 1]))
      CONF_ERROR(cmd,
        "please begin and end hours argument with a number");
  }

  if ((error = is_valid_hours_arg((cmd->argv)[2],
      named_time)) != NULL)
    return ERROR_MSG(cmd, NULL, pstrcat(cmd->tmp_pool,
      (cmd->argv)[0], ": ", error, ": '", (cmd->argv)[2],
      "'", NULL));

  /* Check the day argument, allowing for a wildcard argument for the days
   * argument.
   */
  if ((cmd->argv)[3][0] != '*') {
    if ((cmd->argv)[3][0] == '!') {
      if (!isdigit((int) (cmd->argv)[3][1]) ||
          !isdigit((int) (cmd->argv)[3][strlen((cmd->argv)[3]) - 1]))
        CONF_ERROR(cmd,
          "please begin and end days argument with a number");

    } else {

      if (!isdigit((int) (cmd->argv)[3][0]) ||
          !isdigit((int) (cmd->argv)[3][strlen((cmd->argv)[3]) - 1]))
        CONF_ERROR(cmd,
          "please begin and end days argument with a number");
    }
  }

  if ((error = is_valid_days_arg((cmd->argv)[3],
      named_time)) != NULL)
    return ERROR_MSG(cmd, NULL, pstrcat(cmd->tmp_pool,
      (cmd->argv)[0], ": ", error, ": '", (cmd->argv)[3],
      "'", NULL));

  named_time->name = pstrdup(permanent_pool, (cmd->argv)[1]);

  if (cmd->argc == 5) {
    struct stat statbuf;

    /* The admin gave a filename for a denial message.  Make sure that
     * file exists, then copy the filename into the named_time struct.
     */
    if (pr_fsio_stat((cmd->argv)[4], &statbuf) == -1) {
      return ERROR_MSG(cmd, NULL, pstrcat(cmd->tmp_pool,
        (cmd->argv)[0], ": given filename '", (cmd->argv)[4],
        "' does not exist", NULL));

    } else {
      named_time->deny_mesg_file = pstrdup(permanent_pool,
        (cmd->argv)[4]);
    }
  }

  c = add_config_param(cmd->argv[0], 1, (void *) named_time);
  c->flags |= CF_MERGEDOWN;

  return HANDLED(cmd);
}

MODRET set_classcommandtimes(cmd_rec *cmd) {
  pr_time_map_t *time_map = NULL;
  config_rec *c = NULL;
  char *unknown_name = NULL;

  CHECK_ARGS(cmd, 3);

  if ((time_map = get_named_time_map((cmd->argv)[3],
      &unknown_name)) == NULL)
    return ERROR_MSG(cmd, NULL, pstrcat(cmd->tmp_pool,
      (cmd->argv)[0], ": undefined NamedTime '", unknown_name, "'",
      NULL));

  time_map->targets = pstrdup(permanent_pool, (cmd->argv)[1]);

  /* Check the configuration context, based on the commands to be limited. */
  CHECK_CONF(cmd,
    CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL||CONF_ANON|CONF_DIR|CONF_DYNDIR);

  /* Parse the command-expression argument into the time_map. */
  if (!parse_command_arg(cmd->argv[2], time_map))
    return ERROR_MSG(cmd, NULL, pstrcat(cmd->tmp_pool,
      (cmd->argv)[0], ": too many commands given: '",
      (cmd->argv)[2], "'", NULL));

  c = add_config_param(cmd->argv[0], 1, (void *) time_map);
  c->flags |= CF_MERGEDOWN;

  return HANDLED(cmd);
}

MODRET set_commandtimes(cmd_rec *cmd) {
  pr_time_map_t *time_map = NULL;
  config_rec *c = NULL;
  char *unknown_name = NULL;

  CHECK_ARGS(cmd, 2)

  if ((time_map = get_named_time_map((cmd->argv)[2],
      &unknown_name)) == NULL)
    return ERROR_MSG(cmd, NULL, pstrcat(cmd->tmp_pool,
      (cmd->argv)[0], ": undefined NamedTime '", unknown_name, "'",
      NULL));

  time_map->targets = NULL;

  /* Check the configuration context. */
  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL|CONF_ANON|CONF_DIR);

  /* Parse the command-expression argument into the time_map. */
  if (!parse_command_arg(cmd->argv[1], time_map)) 
    return ERROR_MSG(cmd, NULL, pstrcat(cmd->tmp_pool,
      (cmd->argv)[0], ": too many commands given: '",
      (cmd->argv)[1], "'", NULL));

  /* Get a configuration record pointer. */
  c = add_config_param(cmd->argv[0], 1, (void *) time_map);
  c->flags |= CF_MERGEDOWN;

  return HANDLED(cmd);
}

MODRET set_displayafterhours(cmd_rec *cmd) {
  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL|CONF_ANON);

  add_config_param_str(cmd->argv[0], 1, cmd->argv[1]);
  return HANDLED(cmd);
}

MODRET set_groupcommandtimes(cmd_rec *cmd) {
  pr_time_map_t *time_map = NULL;
  config_rec *c = NULL;
  char *unknown_name = NULL;

  CHECK_ARGS(cmd, 3);
 
  if ((time_map = get_named_time_map((cmd->argv)[3],
      &unknown_name)) == NULL)
    return ERROR_MSG(cmd, NULL, pstrcat(cmd->tmp_pool,
      (cmd->argv)[0], ": undefined NamedTime '", unknown_name, "'",
      NULL));
 
  time_map->targets = pstrdup(permanent_pool, (cmd->argv)[1]);

  /* Check the configuration context, based on the commands to be limited. */
  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL|CONF_ANON|CONF_DIR|
    CONF_DYNDIR);

  /* Parse the command-expression argument into the time_map. */
  if (!parse_command_arg(cmd->argv[2], time_map)) 
    return ERROR_MSG(cmd, NULL, pstrcat(cmd->tmp_pool,
      (cmd->argv)[0], ": too many commands given: '",
      (cmd->argv)[2], "'", NULL));

  c = add_config_param(cmd->argv[0], 1, (void *) time_map);
  c->flags |= CF_MERGEDOWN;
  
  return HANDLED(cmd);
}

MODRET set_usercommandtimes(cmd_rec *cmd) {
  config_rec *c = NULL;
  pr_time_map_t *time_map = NULL;
  char *unknown_name = NULL;

  CHECK_ARGS(cmd, 3);
 
  if ((time_map = get_named_time_map((cmd->argv)[3],
      &unknown_name)) == NULL)
    return ERROR_MSG(cmd, NULL, pstrcat(cmd->tmp_pool,
      (cmd->argv)[0], ": undefined NamedTime '", unknown_name, "'",
      NULL));

  time_map->targets = pstrdup(permanent_pool, (cmd->argv)[1]);

  /* Check the configuration context, based on the commands to be limited. */
  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL|CONF_ANON|CONF_DIR|
    CONF_DYNDIR);

  /* Parse the command-expression argument into the time_map. */
  if (!parse_command_arg(cmd->argv[2], time_map)) 
    return ERROR_MSG(cmd, NULL, pstrcat(cmd->tmp_pool,
      (cmd->argv)[0], ": too many commands given: '",
      (cmd->argv)[2], "'", NULL));
 
  c = add_config_param(cmd->argv[0], 1, (void *) time_map);
  c->flags |= CF_MERGEDOWN;

  return HANDLED(cmd);
}

/* Command handlers
 */

MODRET chk_command_times(cmd_rec *cmd) {
  unsigned char *authenticated = NULL;
  char *deny_mesg_file = NULL;
  char *user = NULL;

  /* Set the current_cmd variable. */
  time_curr_cmd = cmd->argv[0];

  /* This a hack so that this function can and will handle _every_ FTP
   * command _except_ USER, PASS, and QUIT.  If this _is_ a USER, PASS,
   * or QUIT command, bow out now.
   */
  if (strcmp((cmd->argv)[0], C_USER) == 0 ||
      strcmp((cmd->argv)[0], C_PASS) == 0 ||
      strcmp((cmd->argv)[0], C_QUIT) == 0)
    return DECLINED(cmd);

  /* Check to see that the user has logged in (configured USER/PASS records)
   * and return the appropriate response if not.
   */
  authenticated = get_param_ptr(cmd->server->conf, "authenticated", FALSE);
  if (!authenticated || *authenticated == FALSE)
    return DECLINED(cmd);

  /* Retrieve the applicable time restriction.  If none are found, let
   * the command pass now.
   */
  if ((deny_mesg_file = get_deny_mesg(cmd)) == NULL) {
    log_debug(DEBUG3, MOD_TIME_VERSION ": %s allowed by configuration",
      cmd->argv[0]);
    return DECLINED(cmd);
  }

  log_debug(DEBUG3, MOD_TIME_VERSION ": %s denied by configuration",
    cmd->argv[0]);

  /* Check for an empty deny_mesg_file string, signalling no message.
   * Deny connection with a 421 error.
   *
   * NOTE: check RFC959 for the appropriate response codes for each
   *  command, as the ftp client may not handle 421 from some command
   *  properly.  Hmmm...it _does_ allow for 421 from all commands...the
   *  reason I brought this up was the functioning of ftp clients in
   *  passive mode, that issue a PASV before each command.  Such clients
   *  hang when the PASV succeeds, but the LIST or NLST, if DIRS is
   *  limited, is blocked.  Perhaps insert a check for the SF_PASSIVE
   *  flag in session.flags (if session.flags & SF_PASSIVE), and end the
   *  passive connection by sending a NULL byte (how? -- src/data.c's
   *  "void data_close(int quiet)", with quiet = TRUE, might work).
   *
   * OK, have scanned through the section in RFC959 regarding FTP
   *  error codes.  It seems that the returned error code should depend
   *  on the limited command.  For example, limited PASV should be
   *  return a 425 error code (421 would close the connection, as if the
   *  server were shutting down, which is not the case; all connection
   *  related commands should get this error code), file commands
   *  (DIRS, READ, WRITE) should get the 450 error code.
   */

  if (strlen(deny_mesg_file) > 0) {
    log_debug(DEBUG4, MOD_TIME_VERSION ": displaying '%s'", deny_mesg_file);

    if (strcmp(cmd->argv[0], C_PASV) == 0 ||
        strcmp(cmd->argv[0], C_PORT) == 0)
      time_display_file(R_425, deny_mesg_file);
    else
      time_display_file(R_450, deny_mesg_file);

    return ERROR(cmd);

  } else {

    char *denymesg = NULL;

    /* Retrieve the current user name */
    user = (char *) get_param_ptr(cmd->server->conf, C_USER, FALSE);

    /* Check for AccessDenyMsg */
    if ((denymesg = (char *) get_param_ptr((session.anon_config ?
        session.anon_config->subset : cmd->server->conf),
        "AccessDenyMsg", FALSE)) != NULL) {
      log_debug(DEBUG5, MOD_TIME_VERSION ": displaying AccessDenyMsg");
      denymesg = sreplace(cmd->tmp_pool, denymesg, "%u", user, NULL);

    } else {
      log_debug(DEBUG5, MOD_TIME_VERSION ": displaying default message");
      denymesg = pstrcat(cmd->tmp_pool, cmd->argv[0],
        " not allowed at this time.", NULL);
    }

    if (strcmp(cmd->argv[0], C_PASV) == 0 ||
        strcmp(cmd->argv[0], C_PORT) == 0)
      return ERROR_MSG(cmd, R_425, denymesg);

    else
      return ERROR_MSG(cmd, R_450, denymesg);
  }

  return DECLINED(cmd);
}

/* NOTE: chk_login_times() is called AFTER successful authentication by
 *  the client.  This function is NOT called if that authentication fails.
 */
MODRET chk_login_times(cmd_rec *cmd) {
  config_rec *current_conf = NULL;
  char *deny_mesg_file = NULL;
  char *user = NULL, *our_name = NULL, *anon_name = NULL;

  /* First, retrieve the logging-in user's name. */
  user = (char *) get_param_ptr(cmd->server->conf, C_USER, FALSE);

  /* Resolve the user enough to find out the proper configuration to be
   * searching...
   */
  current_conf = resolve_user(cmd->pool, &user, &our_name, &anon_name);

  /* Set the current_cmd variable */
  time_curr_cmd = cmd->argv[0];

  /* Now, check for User/Group/Class/global time restrictions for logins.
   * If none are found, let the command pass now.
   */
  if ((deny_mesg_file = get_deny_mesg(cmd)) == NULL) {
    log_debug(DEBUG5, MOD_TIME_VERSION ": login allowed by configuration");
    return DECLINED(cmd);
  }

  log_debug(DEBUG3, MOD_TIME_VERSION ": login denied by configuration");

  /* Check for an empty deny_mesg_file string, signalling no message.
   * Deny connection with a 421 error.
   */

  if (strlen(deny_mesg_file) > 0) {
    time_display_file(R_421, deny_mesg_file);
    return ERROR(cmd);

  } else {
    char *denymesg = NULL;

    /* Check for AccessDenyMsg */
    if ((denymesg = (char *) get_param_ptr((session.anon_config ?
        session.anon_config->subset : cmd->server->conf),
        "AccessDenyMsg", FALSE)) != NULL) {
      denymesg = sreplace(cmd->tmp_pool, denymesg, "%u", user, NULL);
      return ERROR_MSG(cmd, R_421, denymesg);

    } else {
      return ERROR_MSG(cmd, R_421, "Login denied");
    }
  }

  return DECLINED(cmd);
}

/* Module API tables
 */

static conftable time_conftab[] = {
  { "ClassCommandTimes",	set_classcommandtimes,	NULL },
  { "CommandTimes",		set_commandtimes,	NULL },
  { "DisplayAfterHours",	set_displayafterhours,	NULL },
  { "GroupCommandTimes",	set_groupcommandtimes,	NULL },
  { "NamedTime",		add_namedtime,		NULL },
  { "UserCommandTimes",		set_usercommandtimes,	NULL },
  { NULL }
};

static cmdtable time_cmdtab[] = {
  { PRE_CMD,	C_ANY,	G_NONE,	chk_command_times,	FALSE,	FALSE },
  { PRE_CMD,	C_PASS,	G_NONE,	chk_login_times,	FALSE,	FALSE },
  { 0, NULL }
};

module time_module = {
  NULL, NULL,

  /* Module API version 2.0 */
  0x20,

  /* Module name */
  "time",

  /* Module configuration handler table */
  time_conftab,

  /* Module command handler table */
  time_cmdtab,

  /* Module authentication handler table */
  NULL,

  /* Module initialization function */
  NULL,

  /* Session initialization function */
  NULL
};
